#include "resources.h"

#ifndef RT_MANIFEST
#define RT_MANIFEST 24
#endif

#ifndef CREATEPROCESS_MANIFEST_RESOURCE_ID
#define CREATEPROCESS_MANIFEST_RESOURCE_ID 1
#endif

CREATEPROCESS_MANIFEST_RESOURCE_ID RT_MANIFEST "pylauncher.exe.manifest"

ID_OOICON ICON "application.ico"
